.TH netbug 3
.SH NAME
libnetbug \- library to simulate packet loss, misorder, duplication
.SH SYNOPSIS
#include <netbug.h>
.br
extern int _netbug_verbose;
.br
extern int _netbug_seed;
.br
extern int _netbug_recv_delay;
.br
extern double _netbug_recv_dropchance;
.br
extern double _netbug_recv_storechance;
.br
extern double _netbug_recv_dupchance;
.br
extern double _netbug_recv_misorderchance;
.br
extern int _netbug_recv_maxsize;
.br
extern int _netbug_send_delay;
.br
extern double _netbug_send_dropchance;
.br
extern double _netbug_send_storechance;
.br
extern double _netbug_send_dupchance;
.br
extern double _netbug_send_misorderchance;
.br
extern int _netbug_send_maxsize;

extern int _netbug_recvd;
.br
extern int _netbug_recv_delayed;
.br
extern int _netbug_recv_dropped;
.br
extern int _netbug_recv_stored;
.br
extern int _netbug_recv_duped;
.br
extern int _netbug_recv_misordered;
.br
extern int _netbug_recvd_maxsize
.br
extern int _netbug_sent;
.br
extern int _netbug_send_delayed;
.br
extern int _netbug_send_dropped;
.br
extern int _netbug_send_stored;
.br
extern int _netbug_send_duped;
.br
extern int _netbug_send_misordered;
.br
extern int _netbug_sent_maxsize;

extern void _netbug_init();
extern void _netbug_stats();

.SH DESCRIPTIPON
The
.I netbug
library simulates network errors like packet loss, packet delivery in
misorder, packet duplication, to test the influence of such errors on
functionality and performance of overlying communication
protocols. Since this is a library, it can only be used on user-space
implementations of transport protocols. Another feature is to drop
every packet greater than a given size.

The
.I netbug
library contains replacements for the
.I close(2), read(2), recv(2), recvfrom(2), recvmsg(2), send(2), sendmsg(2), sendto(2), write(2)
system call functions, which will delay, drop, duplicate or deliver in misorder
messages at random. It will disturb only operations on AF_INET SOCK_DGRAM
or SOCK_RAW sockets (i.e. UDP messages or raw IP), and tries to not
interfere whith using system services (YP resp. NIS, DNS, NFS).

The delay values denote the upper bound of uniformly distributed
delays for the system calls (in milli seconds).

The
.I netbug
library needs to linked into the application before the standard C library.
The random chances can be set explicitely by a 
.I netbug
\- aware program via the above listed variables, or can be set via
environment variables. This way is completely transparent to the
application program, it only needs to be linked with the
.I netbug
library. The default behaviour is not to disturb socket operations at all.

_netbug_seed resp. NETBUG_SEED gives the random seed for the disturbance
chances. \-1 means that the return value of
.I time(3)
is used as random seed.

The library counts the number of messages sent, dropped, duplicated
in the global integer variables.
.SH ENVIRONMENT
NETBUG_VERBOSE
.br
NETBUG_SEED
.br
NETBUG_RECV_DELAY
.br
NETBUG_RECV_DROPCHANCE
.br
NETBUG_RECV_STORECHANCE
.br
NETBUG_RECV_MISORDERCHANCE
.br
NETBUG_RECV_DUPCHANCE
.br
NETBUG_RECV_MAXSIZE
.br
NETBUG_SEND_DELAY
.br
NETBUG_SEND_DROPCHANCE
.br
NETBUG_SEND_STORECHANCE
.br
NETBUG_SEND_MISORDERCHANCE
.br
NETBUG_SEND_DUPCHANCE
.br 
NETBUG_SEND_MAXSIZE
.SH FILES
.I libnetbug.a
is the static version of the library,
.I libnetbug.so.*.*
is the shared version of the library.
.SH BUGS
The application will get seriously confused, if it uses
.I select(2)
to wait for incoming packets, but the
.I recv*()
function decides to drop the packet and wait for the next packet.
.br
Workaround: For such an application, you must disturb the sending functions
only. 

The
.I netbug
library uses
.I srandom()
and
.I random(),
which might confuse an application that also uses these functions,
because the application will se different random number sequences than when
linked with the unmodified system library.

.I recvmsg()
Can currently not stored received packets for later delivery, but only
drop packets.

.SH AUTHORS
Stefan Petri <petri@ibr.cs.tu-bs.de>; with features added by Ralph Wittmann
<wittmann@ibr.cs.tu-bs.de>
.SH SEE ALSO
.I close(2), read(2), recv(2), recvfrom(2), send(2), sendto(2),
.I write(2)
.br
.I socket(2), ip(4), udp(4), tcp(4)
.br
.I ld(1)
