/*
 *	pingpong.c
 *
 * test program, that sends datagrams between instantions of itself.
 * Invocation:
 *	pingpong portnumber hostname ...
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <strings.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>


/*#define TIME_CHECKPOINTING*/

#ifdef TIME_CHECKPOINTING
/*
 * --> you really dont want to read the next lines. just skip until the next
 * --> #endif TIME_CHECKPOINTING
 * I dont want to put this section inside a fancy configure thingy.
 * If it doesnt work, just #undef the symbol and ignore it.
 */
#include <sys/times.h>
/*#include <sys/resource.h>*/

#define RUSAGE_TIME(t) ( (t).tv_sec * 1000 + (t).tv_usec / 1000 )

extern int getrusage _Proto((int who, struct rusage *rusage));

static struct rusage ru1, ru2;

/*static struct tms tm1, tm2, tm3, tm4, tm5, tm6, tm7, tm10, tm11, tm12, tm13;*/
/*static clock_t c1, c2, c3, c4, c5, c6, c7, c10, c11, c12, c13;*/
static struct timeval tv1, tv2;
#define MEASURE_RUSAGE(ru, c, tm, tv) \
{ \
  getrusage(RUSAGE_SELF, &ru); \
  /*c = times(&tm);*/ \
  gettimeofday(&tv, NULL); \
}
#define PRINT_RUSAGE(ru, c, tm, tv, str) \
{ \
  unsigned long ut, st; \
  ut = RUSAGE_TIME(ru.ru_utime); \
  st = RUSAGE_TIME(ru.ru_stime); \
  fprintf(stderr, \
	  "Probe: %s tod %ld.%06ld ru_utime %ld.%06ld ru_stime %ld.%06ld minflt %ld majflt %ld nswap %ld inblock %ld oublock %ld nvcsw %ld nivcsw %ld\n", \
	 /* "%s: tod %ld.%06ld clock %f, tms_utime %f tms_stime %f ru_utime %ld.%06ld ru_stime %ld.%06ld minflt %ld majflt %ld nswap %ld inblock %ld oublock %ld msgsnd %ld msgrcv %ld nvcsw %ld nivcsw %ld\n", */\
	  str, (long)tv.tv_sec, (long)tv.tv_usec, \
	  /*(double)c/HZ, (double)tm.tms_utime/HZ, (double)tm.tms_stime/HZ, */\
	  (long)ru.ru_utime.tv_sec, (long)ru.ru_utime.tv_usec, \
	  (long)ru.ru_stime.tv_sec, (long)ru.ru_stime.tv_usec, \
	  (long)ru.ru_minflt, (long)ru.ru_majflt, (long)ru.ru_nswap, \
	  (long)ru.ru_inblock, (long)ru.ru_oublock, \
	  /*(long)ru.ru_msgsnd, (long)ru.ru_msgrcv, */ \
	  (long)ru.ru_nvcsw, (long)ru.ru_nivcsw); \
}
#else
#define MEASURE_RUSAGE(ru, c, tm, tv) { /* empty */ }
#define PRINT_RUSAGE(ru, c, tm, tv, s) { /* empty */ }
#endif /* TIME_CHECKPOINTING */


int ack(x, y) int x, y; {
  if (0 == x) return y + 1;
  if (0 == y) return ack(x - 1, 1);
  return ack(x - 1, ack(x, y - 1));
}


#define ITER 10000

int main(argc, argv)
int argc;
char **argv;
{
  int i, sock, iter;
  struct sockaddr_in addr;
  char buf[256];
  int bufused;

  if (argc < 2) {
    fprintf(stderr, "Usage: %s portnumber [host ...]\n", *argv);
    exit(1);
  }
  if (gethostname(buf, sizeof buf - 1)) strcpy(buf, "amnesiac");
  bufused = strlen(buf)+1;

  addr.sin_family = AF_INET;
  addr.sin_port = atoi(argv[1]);
  addr.sin_addr.s_addr = INADDR_ANY;
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0))<0)
  {
  	perror("cant create socket");
	exit(1);
  }
  if(bind(sock, (struct sockaddr *)&addr, sizeof addr)) {
    perror("cant bind socket");
    exit(1);
  }
  /* now ping a datagram to every given host */
  for (i = 2; i < argc; i++) {
    struct hostent *h;
    h = gethostbyname(argv[i]);
    if (!h) {
      fprintf(stderr, "cant get address for %s - ignored\n", argv[i]);
      continue;
    }
    addr.sin_addr.s_addr = *(int *)h->h_addr_list[0];
    printf("sending to %s 0x%x\n", argv[i], addr.sin_addr.s_addr);
    if (sendto(sock, buf, bufused, 0, (struct sockaddr *)&addr, sizeof addr)
	!= bufused) {
      fprintf(stderr, "cant sendto %s", argv[1]);
      perror("");
    }
  }

  /* now pong the replies */
  MEASURE_RUSAGE(ru1, c1, tm1, tv1);
  for (iter = ITER; iter; iter--) {
    bufused = sizeof buf - 1;
    i = sizeof addr;
    if ((bufused = recvfrom(sock, buf, bufused, 0,
			    (struct sockaddr *)&addr, &i)) <= 0) {
      perror("recvfrom");
      exit(1);
    }
    /* printf("got %d bytes from 0x%x: `%s' - will send back\n",
     *        bufused, addr.sin_addr.s_addr, buf);
     */
    /*ack(3, 4);*/
    if (sendto(sock, buf, bufused, 0, (struct sockaddr *)&addr, sizeof addr)
	!= bufused) {
      fprintf(stderr, "cant sendto 0x%x", addr.sin_addr.s_addr);
      perror("");
    }
  }
  MEASURE_RUSAGE(ru2, c2, tm2, tv2);
  PRINT_RUSAGE(ru1, c1, tm1, tv1, "1");
  PRINT_RUSAGE(ru2, c2, tm2, tv2, "2");
}
