#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "video.h"
#include "avformat.h"
#include "rblock.h"

#define DEBUG 1

/* globals */
char *cstrMovieName;
int nFilenr;
FILE *actualFile;

int _readVBlock(int, char *, t_rblock *);
int _readInt(int, char *, int *);


bool openNextFile(){
	char buf[64];
	snprintf(buf, 64, "%s/%s.%04d",MOVIE_DIRECTORY, cstrMovieName, nFilenr);
	actualFile = (FILE *)fopen(buf,"rb");
 	if(actualFile==NULL){
		return false;
	}
  	if(DEBUG) printf("*** Open File: %s\n",buf);
	return true;
}

int readHeader(char *chunk, t_avfileheader **prHeader){
	*prHeader = (t_avfileheader *)malloc(AVFILEHEADER_SIZE);
	**prHeader = *(t_avfileheader *)chunk;
	return AVFILEHEADER_SIZE;
}


/* cheating: just use _dump_format from rblock.c */
void _dump_format(FORMAT *avformat);

void dumpHeader(t_avfileheader *prHeader){
  	printf("AVFILEHEADER\n*************\n");
  	printf("msec: %d\n",prHeader->msec);
	_dump_format(&(prHeader->avformat));
}

char *readData(){
	int nChunkSize;
	char *chunk;
	fseek( actualFile, 0L, SEEK_END);
	nChunkSize = (int)ftell(actualFile);
	rewind(actualFile);
	chunk = (char*)malloc(nChunkSize*sizeof(char));
	fread(chunk, nChunkSize, sizeof(char), actualFile);
	fclose(actualFile);
	/*    	printf("data read: %d bytes, 0x%x\n",nChunkSize,nChunkSize); */
	return chunk;
}

/* returns an array of t_rblocks that CAN NOT BE rblock_free()d ! */
int readVBlocks(int off, char *chunk, 
				int nNumOfvblocks, t_rblock **outaVBlocks){
  	int i;
	t_rblock *aVBlocks;

	/*    	printf("read video, off: %d  blockcnt: %d\n",off,nNumOfvblocks); */
	/* array allocieren */ 
  	aVBlocks = (t_rblock *)malloc(nNumOfvblocks*RBLOCK_SIZE_LOCAL);
	*outaVBlocks = aVBlocks;

  	/* read vblocks and set their pointers to the data chunk */ 
	for(i=0; i<nNumOfvblocks; ++i){ 
		off = _readVBlock(off, chunk, &(aVBlocks[i])); 
		/*      		printf("blocknr: %d  off: 0x%x (%d)\n",i,off,off);  */
		/*      		rblock_dump(&(aVBlocks[i]));  */
	}  
	return off;
}

void freeVBlocks(t_rblock *avBlock){
    free(avBlock);
}

int _readVBlock(int off, char *chunk, t_rblock *outprVBlock){
	off = _readInt(off, chunk, &( outprVBlock->header.video.marker ) );
	off = _readInt(off, chunk, &( outprVBlock->header.video.block_num ) );
	off = _readInt(off, chunk, &( outprVBlock->header.video.block_size ) );

	assert(BLOCK_VIDEO == outprVBlock->header.video.marker);

	outprVBlock->filenr = nFilenr;

	outprVBlock->payload_size = (outprVBlock->header.video.block_size);
	outprVBlock->payload = (chunk + off);
	off += outprVBlock->payload_size;
	return off;
}

/* returns a freeable t_rblock* */
int readABlock(int off, char *chunk, t_rblock **outprABlock, int size){
	t_rblock *prABlock;
	int marker;

    prABlock = rblock_new_with_space(BLOCK_AUDIO, (size_t)size);
	prABlock->filenr = nFilenr;

	off = _readInt(off, chunk, &marker);
	if (BLOCK_AUDIO != marker) {
		/* this only happends in the last file */
		free(prABlock->payload);
		prABlock->header.hello.port = AUDIO_EOS_MARKER;
		prABlock->payload_size = 0;
		prABlock->payload = NULL;
		goto out;
	}

	memcpy(prABlock->payload, (chunk + off), size);
	off += size;

out:
	*outprABlock = prABlock;
	return off;
}

int _readInt(int off, char *chunk, int *out){
	int i;
	char *buf = (char *)out;
	for(i=0; i<sizeof(int); ++i){
		buf[i] = chunk[off++];
	}
	return off;
}

void dumpPayload(char *payload, int size){
	int i;
	printf("------------- payload --------------\n");
	for(i=0; i<size; ++i){
  		if(i%16==0) printf("\n");
		if(i%2==0) printf(" ");
		printf("%02x ",(unsigned char)payload[i]);
	}
	printf("\n");
}

/* vi:set tabstop=4 shiftwidth=4 noexpandtab: */
