#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <string.h>
#include <assert.h>
#ifdef TIMING_INFO
#include <time.h>
#include <sys/time.h>
#endif
#include "video.h"
#include "avformat.h"
#include "rblock.h"

#define BUFFERSIZE 80 /* video subframe ringbuffer, minimal 1, maximal 80, see spec */
#define INITIALFETCH 45 /* file-data to fetch before start of playback */
#define SECURITYSPACE 25 /* do not fill buffer until it bursts ... */
#define AHEAD_REQUEST 7
#define AUDIOBLOCKSIZE 640 /* size in bytes */
#define SUBFRAMEBUFFERSIZE 15 /* know it's mostly somthing about 6 ... */
#define AUDIOBUFFERSIZE 71 /* audio ring buffer */
#define AUDIOQUEUESIZE 9 /* audio data send to audiodevice buffer */
#define VCLIENT_LOCAL_UDP_PORT 28999

/* globals */
t_rblock audiobuf[AUDIOBUFFERSIZE];
char audiopayloadbuf[AUDIOBLOCKSIZE * AUDIOBUFFERSIZE];
t_rblock subframebuf[SUBFRAMEBUFFERSIZE];
t_rblock *pFormatblk;
t_rblock *pNullaudioblk;
t_rblock *pAckblk;
t_rblock *pReqblk;
int nPlayedAudioBlocks;
int nQueuedAudioBlocks;
int nMaxReceivedAudioBlock;
int nMaxReceivedVideoBlock;
int nLastShownFrame;
int last_req_up_to;

#define MAXINT(a,b)	(a>b ? a : b)
#define BUF_LOW		nPlayedAudioBlocks
#define BUF_HIGH	MAXINT(nMaxReceivedAudioBlock, nMaxReceivedVideoBlock)
#define BUF_MAX		BUF_LOW + BUFFERSIZE

int nStreamFd, nCtrlFd;

#ifdef TIMING_INFO
long lastleavetime;
#endif

/* Prototypen fuer externe Funktionen */
extern void graceful_exit();

enum enClientState{
	INIT = 0,
	FILLBUFFER = 1,
	RECV_AND_PB = 2,
	FINISH = 3,
	QUIT = 4
} enProgress;

void startAudioPB(){
	int i;
	for(i=0; i<AUDIOQUEUESIZE; ++i){
		if(NULL != audiobuf[i].payload){
			play_audio_block((char *)audiobuf[i].payload,
							 audiobuf[i].payload_size);
		}
		else{
			play_audio_block((char *)pNullaudioblk->payload,
							 pNullaudioblk->payload_size);
		}
	}
	printf("started audio pb\n");
	nQueuedAudioBlocks = AUDIOQUEUESIZE;
}

/* gets all subframes beginning with keyframe at pos 0 */
void decompPic(t_rblock *prvblk){
	int nPos, subframenr, check;
	nPos = prvblk->filenr;
	subframenr = prvblk->header.video.block_num % (BUFFERSIZE*pFormatblk->header.format.avformat.block_anz);
/*  	printf("subframe: %d  size: %d  offset: %d\n",   */
/*  		   subframenr, prvblk->header.video.block_size,   */
/*  		   (subframenr*pFormatblk->header.format.avformat.bytes_per_block));   */
	
	assert(prvblk->payload != NULL);
	check = decomp_JPEG(prvblk->payload, 
						prvblk->header.video.block_size, 
						(subframenr*pFormatblk->header.format.avformat.bytes_per_block)); 
	/*  		printf("%d ",picCnt); */
	if(0==check){
		printf("JPEG decomp failed! filenr: %d  picnr: %d",
			   nPos, subframenr);
	}
}

void queueAudio(int nPos){
	int apos = (nPos%AUDIOBUFFERSIZE);
	/* if audio block outdated add nullblock */
	if(NULL==audiobuf[apos].payload || nPos != audiobuf[apos].filenr){
		play_audio_block((char *)pNullaudioblk->payload, pNullaudioblk->payload_size);
	}
	else{
		if (AUDIO_EOS_MARKER == audiobuf[apos].header.hello.port) {
			goto end_of_stream_reached;
		}
		play_audio_block((char *)audiobuf[apos].payload, audiobuf[apos].payload_size);
	}
	++nQueuedAudioBlocks;
	return;

end_of_stream_reached:
	/* queue a final null audio block (played while showing last frame): */
	/* (disabled because it produces a nasty "click"-sound - better to */
	/* just skip the last frame) */
	/* play_audio_block((char *)pNullaudioblk->payload, pNullaudioblk->payload_size); */
	enProgress = FINISH;
	return;
}

void showVideo(int nPos){
	printf("video: %d (skipped: %d)\n",nPos, (nPos-nLastShownFrame-1));
	show_video((nPos%BUFFERSIZE)*pFormatblk->header.format.avformat.height);
	nLastShownFrame = nPos;
}

void feierabend()
{
	rblock_free(pFormatblk);
	rblock_free(pAckblk);
	rblock_free(pReqblk);
	rblock_free(pNullaudioblk);
	graceful_exit();
}

void bufferLogo(){
	int nChunkSize;
	char *fname = "avsynclogo.jpg";
	char *chunk;
	FILE *logoFile;
	logoFile = (FILE *)fopen(fname,"rb");
 	if(logoFile!=NULL){
		printf("logo opened ;-).\n");
		fseek( logoFile, 0L, SEEK_END);
		nChunkSize = (int)ftell(logoFile);
		rewind(logoFile);
		chunk = (char*)malloc(nChunkSize*sizeof(char));
		fread(chunk, nChunkSize, sizeof(char), logoFile);
		fclose(logoFile);
		decomp_JPEG(chunk, 320*240*3, 0);
		show_video(0);
		free(chunk);
	}
}

void init(char *hostName, int portNr, char *movie)
{
	int i;
	/*int minbufsize = 32 * 1024;
	  int bufsize = 256 * 1024;
	  int ret;*/
	int s;
	int ss;
	struct sockaddr_in streamaddr;
	int yes = 1;

    struct addrinfo hints;
    struct addrinfo *res=NULL;
	char cstrport[15];

	t_rblock* helloblk;

	printf("Playing %s@%s:%d...\n",movie,hostName,portNr);

	/* open local UDP port */
	nStreamFd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if(nStreamFd < 0) { 
		perror("Error opening stream socket");
		exit(1);
	}
	/* increase receive buffer
	   do {
	   ret = setsockopt(nStreamFd, SOL_SOCKET, SO_RCVBUF,
	   (char *)&bufsize, sizeof bufsize);
	   if (ret < 0)
	   bufsize /= 2;
	   } while (ret < 0 && bufsize >= minbufsize);*/
	ss = sizeof s;
	if (getsockopt(nStreamFd, SOL_SOCKET, SO_RCVBUF, (char*) &s, &ss) < 0)
		perror("getsockopt(SO_RCVBUF)");
	else printf("socket receive buffer is %d bytes.\n", s);
  
	streamaddr.sin_family=AF_INET;
	streamaddr.sin_addr.s_addr=INADDR_ANY;
	streamaddr.sin_port=VCLIENT_LOCAL_UDP_PORT;

	if(bind(nStreamFd, (struct sockaddr*)&streamaddr, sizeof(streamaddr)) < 0) {
		perror("Error binding local stream address");
		exit (1);
	}

	/* bind to server TCP port */
	sprintf(cstrport,"%d",portNr);

	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	hints.ai_flags = AI_PASSIVE;

	if(0!=getaddrinfo( hostName, cstrport, &hints, &res )){
		printf( "Failed to resolve name or address.\n");
		exit(1);
	}

	if((nCtrlFd  = socket ( res->ai_family, res->ai_socktype, 0)) < 0) { 
		perror("Error opening socket");
		exit(1);
	}

	/* set TCP_NODELAY */
	if(-1 == setsockopt(nCtrlFd, SOL_TCP, TCP_NODELAY, &yes, sizeof(yes))) {
		perror ("Error setting TCP control socket option TCP_NODELAY");
		exit (1);
	}
	  
	/* connect to the host */    
	if(connect(nCtrlFd, res->ai_addr, res->ai_addrlen) < 0) {
		perror("Error connecting host");
		exit(1);
	}

	/* send hello block */
	helloblk = rblock_new(BLOCK_HELLO);
	helloblk->header.hello.port = VCLIENT_LOCAL_UDP_PORT;
	helloblk->payload = strdup(movie);
	helloblk->payload_size = strlen(movie);

	rblock_send(nCtrlFd, helloblk, 0);
	rblock_free(helloblk);

	/* get header block */
	pFormatblk = rblock_recv(nCtrlFd, 0);
	if((NULL == pFormatblk) || (pFormatblk->header.marker != BLOCK_FORMAT)){
		printf("Error getting format info.");
		exit(1);
	}

	/* init globals */
	for(i=0; i<AUDIOBUFFERSIZE; ++i){
		audiobuf[i].filenr = -1;
		audiobuf[i].payload = NULL;
	}
	nQueuedAudioBlocks = 0;
	nMaxReceivedAudioBlock = 0;
	nMaxReceivedVideoBlock = 0;
	last_req_up_to = AHEAD_REQUEST;

	/* init global blocks */
	pNullaudioblk = rblock_new_with_space(BLOCK_AUDIO, AUDIOBLOCKSIZE);
	pReqblk = rblock_new(BLOCK_REQ);
	pAckblk = rblock_new(BLOCK_AUDIO_ACK);

	enProgress = INIT;
}

int inline calc_req_up_to(int i){
	int req_up_to;

	/* assert we still have some space: */
	/* (XXX: assert for video and audio seperately ? )*/
	assert(BUF_HIGH < BUF_MAX);

#ifndef PHILIPP_ALGO
	if(FILLBUFFER==enProgress){
		req_up_to = last_req_up_to + i;
		if(req_up_to>(BUF_HIGH + AHEAD_REQUEST))
			req_up_to = BUF_HIGH + AHEAD_REQUEST;
	}
	else{
		req_up_to = last_req_up_to + i;
		if(req_up_to>(BUF_HIGH + INITIALFETCH + AHEAD_REQUEST)) 
			req_up_to = BUF_HIGH + INITIALFETCH + AHEAD_REQUEST;
		if(req_up_to>(nPlayedAudioBlocks+INITIALFETCH)) 
			req_up_to = nPlayedAudioBlocks + INITIALFETCH + AHEAD_REQUEST;
	}

	return req_up_to;
#else
	switch (enProgress) {
		case INIT:
		case FILLBUFFER:
			return BUF_HIGH + 1;
		default:
			return BUF_MAX - SECURITYSPACE;
	}
#endif
}

void requestData(int i){
	int req_up_to;

	req_up_to = calc_req_up_to(i);
	if(req_up_to > last_req_up_to){
		pReqblk->header.status.send_up_to = req_up_to;
		rblock_send(nCtrlFd, pReqblk, 0);
		printf(" send_up_to:%d ", req_up_to);
	}
	last_req_up_to = req_up_to;
}

void ack_audio_block(int nr) {
	/* calc the maximum file number to request */

	pAckblk->filenr = nr;
	pAckblk->header.status.send_up_to = last_req_up_to;
	pAckblk->header.status.acked_file = nr;
	rblock_send(nCtrlFd, pAckblk, 0);
}

bool fetchData(){
	int apos;
	t_rblock recvblk;
	struct timeval timeout;
	fd_set fdSet;
	int err;
	char payloadbuf[2048], *pAudioPayload;
	
	FD_ZERO(&fdSet);
	FD_SET(nStreamFd,&fdSet);
	
	/* timeout == 0 : don't wait */
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;

	if(0 > select(nStreamFd+1, &fdSet, NULL, NULL, &timeout)) 
	{
		printf("Error selecting socket");
	}
	if(FD_ISSET(nStreamFd, &fdSet)){
		err = rblock_recv_into(nStreamFd, &recvblk, 0, payloadbuf, 2048);
		if(err <= 0){
			printf("!!! packet recv error !!!\n");
			return false;
		}
		else{
			printf("- Read ");
			rblock_dump(&recvblk);
			switch(recvblk.header.marker){
			case BLOCK_VIDEO:
				if (recvblk.filenr > BUF_LOW+BUFFERSIZE) {
					/* video block too far in the future; discard */
					printf("!!!!! recv video from the future !!!!!\n");
					goto out_ok;
				}
				if (recvblk.filenr < nPlayedAudioBlocks) {
					/* video block too old; discard */
					printf("!!!!! recv outdated video block !!!!!\n");
					goto out_ok;
				}
				if(recvblk.filenr > nMaxReceivedVideoBlock){
					nMaxReceivedVideoBlock = recvblk.filenr;
				}

				decompPic(&recvblk);
				goto out_ok;
			case BLOCK_AUDIO:
				if (recvblk.filenr > BUF_LOW+AUDIOBUFFERSIZE) {
					/* audio block too far in the future;
					 * discard and hope for resend */
					printf("!!!!! recv audio from the future !!!!!\n");
					goto out_ok;
				}

				if (recvblk.filenr < nQueuedAudioBlocks) {
					/* audio block too old; ack & discard */
					printf("!!!!! recv outdated audio block !!!!!\n");
  					ack_audio_block(recvblk.filenr);
					goto out_ok;
				}
				if(recvblk.filenr > nMaxReceivedAudioBlock){
					nMaxReceivedAudioBlock = recvblk.filenr;
				}

				apos = ((recvblk.filenr)%AUDIOBUFFERSIZE);
				rblock_copy(&recvblk, &(audiobuf[apos]));
				pAudioPayload = audiopayloadbuf + apos*AUDIOBLOCKSIZE;
				memcpy(pAudioPayload,
					   payloadbuf,
					   audiobuf[apos].payload_size);
				audiobuf[apos].payload = pAudioPayload;
				ack_audio_block(recvblk.filenr);

				/* don't free payload (still a copy to it), just return */
				goto out_ok_wo_free;
			case BLOCK_FORMAT:
			case BLOCK_HELLO:
			case BLOCK_AUDIO_ACK:
			case BLOCK_REQ:
			default:
				printf("What the heck?! Got an unexpected packet...!\n");
			}
			goto out_ok;
		}
	}
	return false;
	
out_ok:
	if (NULL != recvblk.payload) {
/*  		free(recvblk.payload); */
	}
out_ok_wo_free:
	return true;
}


/* Die Funktion main_loop wird periodisch vom Hauptprogramm durchlaufen.     */
/* In dieser Funktion werden die Daten des Servers empfangen und bearbeitet. */
void main_loop()
{
  	int i;

#ifdef TIMING_INFO
	struct timeval rNow;
	long entertime, leavetime;

	if (0>gettimeofday(&rNow, NULL)) {
		perror("gettimeofday");
		exit(1);
	}
	entertime = rNow.tv_sec * 1000 + rNow.tv_usec / 1000;
#endif

	switch(enProgress){
	case RECV_AND_PB:
		nPlayedAudioBlocks = get_completed_audio_blocks();
		/* request fresh data */
		requestData(1);

		/* fill buffers with received data */
		for(i=0; i<10; ++i) {
			if (!fetchData()) break;
		}

		/* play outstanding av data */
		nPlayedAudioBlocks = get_completed_audio_blocks();
		printf("blow:%d bhigh:%d mrab:%d mrvb:%d qab:%d pab:%d ",BUF_LOW, BUF_HIGH, nMaxReceivedAudioBlock, nMaxReceivedVideoBlock, nQueuedAudioBlocks, nPlayedAudioBlocks);

		/* play next audio block (increments nQueuedAudioBlocks */
		queueAudio(nQueuedAudioBlocks+1);
		showVideo(nPlayedAudioBlocks+1);

		/*
		a2queue = AUDIOQUEUESIZE - (nQueuedAudioBlocks - nPlayedAudioBlocks);
		printf(" a2q:%d\n",a2queue);
		for(i=0; i<a2queue; ++i) {
			assert(a2queue>0);
			queueAudio(nQueuedAudioBlocks+1);
		}
		if (nLastShownFrame < nPlayedAudioBlocks+1) {
			showVideo(nPlayedAudioBlocks+1);
		}
		*/
		break;
	case FILLBUFFER:
		requestData(1);
		for(i=0; i<7; ++i) {
			fetchData();
		}
		printf("%d files fetched. ",BUF_HIGH);
		if(INITIALFETCH <= BUF_HIGH){
			startAudioPB();
			enProgress = RECV_AND_PB;
		}
		break;
	case INIT:
		printf("av init, video BUFFERSIZE: %d\n",BUFFERSIZE);  
		init_audio_video(&(pFormatblk->header.format.avformat), BUFFERSIZE); 
		printf("initialized.\n");  
		enProgress = FILLBUFFER;
		bufferLogo();
		break;
	case FINISH:
		nPlayedAudioBlocks = get_completed_audio_blocks();

		if (nPlayedAudioBlocks >= nQueuedAudioBlocks) {
			enProgress = QUIT;
		}

		/*show next video block: */
		showVideo(nPlayedAudioBlocks+1);
		break;
	case QUIT:
		feierabend();
		exit(1);
	}

#ifdef TIMING_INFO
	if (0>gettimeofday(&rNow, NULL)) {
		perror("gettimeofday");
		exit(1);
	}
	leavetime = rNow.tv_sec * 1000 + rNow.tv_usec / 1000;

	printf("// leaving mainloop, t_out: %ld, t_in: %ld\n",
			entertime-lastleavetime,
			leavetime-entertime);
	lastleavetime=leavetime;
#endif
}

/* vi:set tabstop=4 shiftwidth=4 noexpandtab: */
