#define _GNU_SOURCE
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>

#include "video.h"
#include "avformat.h"
#include "rblock.h"
#include "reader.h"

#define FRAME_TTL 300 /* in msec*/

/* globals */
int nStreamFd, nCtrlFd, nListenFd;
int nSendUpTo, nHighestFileSent;

typedef struct rlinkedlist{
	t_rblock *block;
	int timeout;
	struct rlinkedlist *next;
} t_rlinkedlist;

t_rlinkedlist *llist_first = NULL;
t_rlinkedlist *llist_last = NULL;

void _append(t_rlinkedlist *this) {
	this->next = NULL;
	if(NULL != llist_first){
		llist_last->next = this;
		llist_last = this;
	}
	else{
		llist_first = this;
		llist_last = llist_first;
	}
}

void _unlink(t_rlinkedlist *prev, t_rlinkedlist *this) {
	if(prev!=NULL){
		prev->next = this->next;
	} else
	if (this == llist_first) {
		llist_first = this->next;
		if (NULL == llist_first) {
			llist_last = NULL;
		}
	} else {
		printf("error in linked list managment\n");
		exit(1);
	}
	if (llist_last == this) {
		llist_last = prev;
	}
	this->next = NULL;
}

void appendList(t_rblock *rBlock2Add, long timeNow){
	t_rlinkedlist *prNewFrame;
	prNewFrame = (t_rlinkedlist *)malloc(sizeof(t_rlinkedlist));
	prNewFrame->block = rBlock2Add;
	prNewFrame->timeout = (timeNow + FRAME_TTL);
	_append(prNewFrame);
}

int removeAudioBlock(int nBlockNr){
	if(NULL != llist_first){
		t_rlinkedlist *iter = llist_first;
		t_rlinkedlist *prev = NULL;
		while(NULL != iter){
			if(nBlockNr == iter->block->filenr){
				_unlink(prev, iter);
				rblock_free(iter->block);
				free(iter);
				return true;
			}
			prev = iter;
			iter = iter->next;
		}
	}
	return false;
}

int checkTimeoutsAndResend(long now){
	t_rlinkedlist *iter = llist_first;
	t_rlinkedlist *prev = NULL;
	int nCnt = 0;
	if(NULL == iter){
		printf("check timeout: Audio Buffer run empty!!!\n");
		return 0;
	}
	else{
		while(NULL != iter
			  &&
			  now > iter->timeout)
		{
			_unlink(prev, iter);
			rblock_send(nStreamFd, iter->block, 0);
			iter->timeout = (now + FRAME_TTL);
			_append(iter);

			prev = iter;
			iter = iter->next;
			++nCnt;
		}
	}
	return nCnt;
}

/******************************************************************************/

void send_file(long now){
  	char *chunk;
  	int off;
	int picCnt;
  	t_avfileheader *prHeader;
  	t_rblock *aVBlocks;
  	t_rblock *prABlock;
  	t_rblock *formatblk;

  	if(openNextFile()==FALSE){
  		/* fertich :-) */
		return;
  	}

	/*    	printf("starting with file: %d\n",nFilenr); */
  	chunk = readData();
  	off = 0;
  	off = readHeader(chunk, &prHeader);
	if(0==nFilenr){ 
		formatblk = rblock_new(BLOCK_FORMAT);
		memcpy( &(formatblk->header.format.avformat),
				&(prHeader->avformat),
				sizeof(FORMAT));
		
		rblock_dump(formatblk);
		rblock_send(nCtrlFd,formatblk,0);
	}  
	/*  	printf("read video, off: 0x%x (%d)  blockcnt: %d\n",off,off, */
	/*  		   prHeader->avformat.block_anz); */
	off = readVBlocks(off, chunk, prHeader->avformat.block_anz, &aVBlocks);
	/*  	printf("video read, off: 0x%x (%d)\n",off,off);  */
	off = readABlock(off, chunk, &prABlock,  
  					 prHeader->avformat.audio_buffer_size); 
	/*  	printf("audio read, off: 0x%x (%d)\n",off,off);  */


	rblock_send(nStreamFd, prABlock, 0);
	appendList(prABlock, now);

	for(picCnt=0; picCnt<prHeader->avformat.block_anz; ++picCnt){
		rblock_send(nStreamFd, &(aVBlocks[picCnt]), 0);
	}

	free(chunk);
	free(aVBlocks);

	if (nFilenr > nHighestFileSent) {
		nHighestFileSent = nFilenr;
	}
	++nFilenr;
}

void mainloop(){
	t_rblock* blk;
	struct timeval timeout, rNow;
	long now;
	fd_set fdSet;

	/* 1. check for timeouts on unacknowledged packets */
	if (0>gettimeofday(&rNow, NULL)) {
		perror("gettimeofday");
		exit(1);
	}
	now = rNow.tv_sec * 1000 + rNow.tv_usec / 1000;
	checkTimeoutsAndResend(now);

	/* 2. send out as many files as i'm allowed */
	if (nSendUpTo > nHighestFileSent) {
		send_file(now);
	}

	/* 3. check control socket for ACKs/Status */
	FD_ZERO(&fdSet);
	FD_SET(nCtrlFd,&fdSet);

	/* short timeout, so we can timeout audio blocks */
	timeout.tv_sec = 0;
	timeout.tv_usec = 10000;

	if(0 > select(nCtrlFd+1, &fdSet, NULL, NULL, &timeout)) 
	{
		printf("select error");
	}
	if(FD_ISSET(nCtrlFd, &fdSet)){
		blk = rblock_recv(nCtrlFd,0);
		if (blk == NULL) {
			/* error occured or stream closed by client */
			perror("bye...");
			exit(0);
		}
		printf("Received ");
		rblock_dump(blk);

		switch (blk->header.marker) {
			case BLOCK_AUDIO_ACK:
				removeAudioBlock(blk->header.status.acked_file);

				/* fall through */
			case BLOCK_REQ:
				nSendUpTo = blk->header.status.send_up_to;
				break;

			case BLOCK_VIDEO:
			case BLOCK_AUDIO:
			case BLOCK_HELLO:
			case BLOCK_FORMAT:
			default:
				perror("weird stuff happened !");
				exit(1);
		}
	}
}

int main(int argc, char *argv[])
{
	int nListenPort, nTargetPort;
	struct sockaddr_in rServerAddr, rClientCtrlSrc, rClientStreamTarget;
	int addrlen = sizeof (struct sockaddr_in);
	t_rblock* helloblk;
	int yes = 1;
	char filename[128];

	if(2 == argc){
		nListenPort = atoi(argv[1]);
	} else {
		printf("Usage: %s [port]\n",argv[0]);
		exit(1);
	}

	/* create, bind and listen to TCP socket */
	rServerAddr.sin_family      = AF_INET;
	rServerAddr.sin_port        = htons(nListenPort);
	rServerAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	nListenFd = socket ( AF_INET, SOCK_STREAM, 0);
	if(-1 == nListenFd){
		perror ("Error creating TCP control socket");
		exit (1);
	}
	if(-1 == setsockopt(nListenFd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes))) {
		/* while debugging, this is really helpful: 
		 * cheat a bit and set SO_REUSEADDR */
		perror ("Error setting TCP control socket option SO_REUSEADDR");
		exit (1);
	}

	if(-1 == bind(nListenFd, (struct sockaddr *)&rServerAddr, sizeof(struct sockaddr))){
		perror ("Error binding TCP control socket");
		exit (1);
	}
	if(-1 == listen (nListenFd, 1)){
		perror("Error listening to TCP control socket");
		exit (1);
	}

	printf ("Server up and waiting for client to connect.\n");

	/* Accept remote host to connect */
	nCtrlFd = accept(nListenFd,(struct sockaddr *) &rClientCtrlSrc, &addrlen);
	if(nCtrlFd == -1){
		perror("Error accepting client connection");
		exit(1);
	}
	printf ("Connection established with %s\n", inet_ntoa(rClientCtrlSrc.sin_addr));

	helloblk = rblock_new(BLOCK_HELLO);
	rblock_recv_into(nCtrlFd, helloblk, 0, filename, 127);
	if (NULL == helloblk) {
		perror("couldn't get HELLO from client");
		exit(1);
	}
	if (helloblk->header.marker != BLOCK_HELLO) {
		perror("protocol error, client didn't send HELLO !");
		exit(1);
	}
  	filename[helloblk->payload_size] = '\0';
	
	cstrMovieName = strdup(filename);
	printf("Client requested movie \"%s\"\n", cstrMovieName);
	
	/* create and bind UDP socket for a/v streaming */
	nTargetPort = helloblk->header.hello.port;
	rClientStreamTarget.sin_family      = AF_INET;
	rClientStreamTarget.sin_port        = htons(nTargetPort);
	rClientStreamTarget.sin_addr.s_addr = rClientCtrlSrc.sin_addr.s_addr;
	printf ("Will stream data to %s:%d\n",
			inet_ntoa(rClientStreamTarget.sin_addr), nTargetPort);

	nStreamFd = socket(AF_INET, SOCK_DGRAM, 0);
	if(-1 == nStreamFd){
		perror("Error creating streaming socket");
		exit(1);
	}
	if (-1 == connect(nStreamFd, (struct sockaddr*)&rClientStreamTarget,
				sizeof(rClientStreamTarget))) {
		perror("connect for UDP socket failed !?");
		exit(1);
	}

	/* allow to send some files immediately */
	nSendUpTo = 1;

	/* Enter mainlopp */
	while(TRUE) {
		mainloop();
	}

	return 0;
}

/* vi:set tabstop=4 shiftwidth=4 noexpandtab: */
