/*===========================================================
 FILENAME     avformat.h
 VERSION      0.1

 DESCRIPTION  Transport Format for Audio and Video Stream
 AUTHOR       pks05

 CHANGES
 Date       | Name             | Comment
 ------------------------------------------------------------
 2004-07-09 | Sonnenberg       | Initial Version
 
 vi:set tabstop=4 shiftwidth=4 noexpandtab:
=============================================================*/
#include "video.h"

#ifndef __AVFORMAT_H
#define __AVFORMAT_H

/*----- General defines -----*/
#define AV_MAXBUFFSIZE        4096
#define AV_MAXSTRSIZE         80

#ifndef bool
typedef enum enBool { false = 0, true = 1 } bool;
#endif /* bool */

/*--------------------------*/
/*----- AVFormat Types -----*/
/*--------------------------*/

typedef enum enBlockMarker
{
	BLOCK_VIDEO		= DGRAM_VIDEO,
	BLOCK_AUDIO		= DGRAM_AUDIO,
	BLOCK_FORMAT	= DGRAM_INFO,
	BLOCK_HELLO		= 15,
	BLOCK_AUDIO_ACK	= 16,
	BLOCK_REQ		= 17
} t_enBlockMarker;

/* Bildheader */
typedef struct avfileheader
{
	/* Zeit seit Start des Filmes */
	/* in Millisekunden */
	/* Format-Block */
	int msec;
	/* avformat, s. video.h */
	FORMAT avformat;
} t_avfileheader;

#define AVFILEHEADER_SIZE		sizeof(t_avfileheader)

/* Traeger fuer einen Audio- oder Video-Block */
typedef struct rblock
{
	/* number of file / chunk */
	int filenr;

	/* time position of block in msec */
	int msec;

	union {
		/* Video Block Header, s. video.h */
		video_h				video;

		/* oder Audio Block Header */
		audio_h				audio;

		/* oder FORMAT Infos */
		struct rFormatHeader {
			int				marker;
			FORMAT			avformat;
		} format;

		/* oder einen freundlichen Gruss */
		struct rHelloHeader {
				int			marker;
				int			port;
				/* filename in payload */
		}					hello;

		/* benutzt fr BLOCK_AUDIO_ACK und BLOCK_REQ */
		struct rStatusHeader {
			int				marker;
			int				send_up_to;
			int				acked_file;	/* only valid in BLOCK_AUDIO_ACK */
		} status;

		/* oder nur direkter Zugriff auf den Marker */
		t_enBlockMarker		marker;
	} header;

	/* size of the payload in octets */
	int			payload_size;

	/* The payload is referenced when in-memory,
	 * and just appended during transmission */
	void *		payload;

} t_rblock;

#define RBLOCK_SIZE_LOCAL		sizeof(t_rblock)
#define RBLOCK_SIZE_STREAMED	(sizeof(t_rblock) - sizeof(void*))

#define AUDIO_EOS_MARKER		0x0DEADBABE
#endif /* __AVFORMAT_H */
